/**
 * Fractal Binding Factory.
 * Copyright (C) 2007-2008 INRIA, SARDES

 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contact: Valerio.Schiavoni@gmail.com
 * 
 * Author: Valerio Schiavoni
 * 
 */
package org.objectweb.fractal.bf.connectors.rmi.asm;

import java.io.BufferedInputStream;
import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.logging.Logger;

/**
 * Instances of this class loader should be obtained using
 * {@link RMIServiceClassLoaderFactory}.
 */
public class RMIServiceClassLoader extends ClassLoader {
	Logger log = java.util.logging.Logger.getLogger(RMIServiceClassLoader.class
			.getCanonicalName());

	/**
	 * Stores the binary names of the classes defined by this class loader and
	 * the corresponding Class and byte objects.
	 */
	private final Map<String, LoadedClass> rmiClasses;

	/**
	 * 
	 * @param parentClassLoader
	 *            the parent class loader for this class loader.
	 */
	public RMIServiceClassLoader(ClassLoader parentClassLoader) {

		super(parentClassLoader);
		log.finest("parent classloader for RMIServiceClassLoader: "
				+ parentClassLoader);
		this.rmiClasses = new HashMap<String, LoadedClass>();
		// is this necessary ?
		Thread.currentThread().setContextClassLoader(this);
	}

	public Class<?> defineClass(byte[] byteArray, String binaryName) {

		log.finest("Binary name for class to define : " + binaryName);
		Class<?> result = null;
		if (!rmiClasses.containsKey(binaryName)) {
			result = super.defineClass(binaryName, byteArray, 0,
					byteArray.length);

			this.rmiClasses.put(binaryName, new LoadedClass(result, byteArray));

		} else {
			result = this.rmiClasses.get(binaryName).clazz;
		}
		log.finest("Class returned by defineClass: " + result);
		return result;
	}

	public Class<?> getClass(String binaryName) {
		return this.rmiClasses.get(inferClassName(binaryName)).clazz;
	}

	@Override
	public InputStream getResourceAsStream(String name) {
		log.finest("Name of resource: " + name + " (classes defined: "
				+ this.rmiClasses + ")");

		InputStream is = null;

		String nameForDefinedClass = inferClassName(name);

		LoadedClass found = this.rmiClasses.get(nameForDefinedClass);
		if (found != null) {
			is = found.toInputStream();
		} else {
			is = super.getResourceAsStream(name);
		}

		return is;
	}

	/**
	 * @param name
	 * @return
	 */
	private String inferClassName(String name) {
		String nameForDefinedClass = name.substring(0, name.lastIndexOf("."))
				.replace("/", ".");
		return nameForDefinedClass;
	}

	/**
	 * Return a view of the binary names for the classes defined by this class
	 * loader. Modifications to this set will NOT be reflected to modifications
	 * to the set of classes defined by this classloader.
	 * 
	 * @returna copy of set of defined classes by this class loader.
	 */
	public Set<String> getDefinedClasses() {
		return new HashSet<String>(rmiClasses.keySet());
	}

	/**
	 * A class to keep track of classes (and their bytes) defined by this
	 * classloader.
	 */
	class LoadedClass {
		private final Class<?> clazz;
		private final byte[] raw;

		LoadedClass(Class<?> c, byte[] raw) {
			this.clazz = c;
			this.raw = raw;
		}

		InputStream toInputStream() {
			return new BufferedInputStream(new ByteArrayInputStream(raw));
		}
	}
}