/**
 * Fractal Binding Factory.
 * Copyright (C) 2007-2008 INRIA, SARDES

 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contact: Valerio.Schiavoni@gmail.com
 * 
 * Author: Valerio Schiavoni
 * 
 */
package org.objectweb.fractal.bf.connectors.rmi.asm;

import java.io.Serializable;
import java.lang.reflect.Method;
import java.rmi.Remote;
import java.rmi.RemoteException;
import java.util.logging.Logger;

import org.objectweb.asm.ClassWriter;
import org.objectweb.asm.Opcodes;
import org.objectweb.asm.Type;

/**
 * A simple class transformer that given a Java interface creates a
 * corresponding Java interface, as a class object, that can be exported via
 * Java RMI. This means that:
 * <ol>
 * <li>the resulting class extends {@link Remote}
 * <li>for every method declared by the interface, a {@link RemoteException} is
 * added to the signature
 * </ol>
 * 
 * The idea behind this code is inspired by a similar one adopted in the
 * Binding-RMI in Apache Tuscany. See:
 * https://svn.apache.org/repos/asf/incubator
 * /tuscany/java/sca/modules/binding-rmi
 * /src/main/java/org/apache/tuscany/sca/binding/rmi/RMIService.java
 * 
 * Their solution adopts a mix of ASM (see http://asm.objectweb.org) and CGLIB
 * bytecode instrumentation tools. In contrast, to reduce the disk footprint for
 * those who wants to use the fractal-bf-connector-rmi plugin, this solution
 * uses ASM only.
 * 
 * @see http://asm.objectweb.org
 * @see http://cglib.sourceforge.net/
 */
public class RemoteInterfaceGenerator {
	/**
	 * The suffix appended to the internal binary name of the generated class.
	 */
	private static final String GENERATED_CLASS_SUFFIX = "$$BindingFactoryGenerated";
	/* The logger to use for this class */
	Logger log = Logger.getLogger(RemoteInterfaceGenerator.class
			.getCanonicalName());

	/**
	 * Create a new java interface that is a loose copy of the input one, by
	 * letting it extends {@link java.rmi.Remote} and by adding a throws clause
	 * to every method so that it throws {@link RemoteException}.
	 * 
	 * @param targetJavaInterface
	 *            the java interface to be transformed
	 * @return a newly generated java interface to be exported with RMI
	 */
	public Class<?> transform(Class<?> targetJavaInterface) {
		final String javaItfCanonicalName = targetJavaInterface
				.getCanonicalName();

		Class<?> toReturn = null;

		if (!targetJavaInterface.isInterface()) {
			throw new IllegalArgumentException("The input class : "
					+ targetJavaInterface.getCanonicalName()
					+ " is not an interface ");
		} else {

			RMIServiceClassLoader classloader = RMIServiceClassLoaderFactory
					.getRMIClassLoader();
			log.finer("JavaRMI class loader ref:  " + classloader);

			final byte[] byteCode = generateRemoteInterface(targetJavaInterface);
			log.finest("Bytecode for RMI-enabled interface generated");
			log.finest("Original java interface name: " + javaItfCanonicalName);
			final String generatedInterfaceInternalName = javaItfCanonicalName
					+ GENERATED_CLASS_SUFFIX;
			log.finest("Generated java interface name: "
					+ generatedInterfaceInternalName);

			Class<?> remotedTargetJavaInterface = classloader.defineClass(
					byteCode, generatedInterfaceInternalName);
			log.finest("Class " + remotedTargetJavaInterface.getCanonicalName()
					+ " for RMI-enabled interface defined in the "
					+ RMIServiceClassLoader.class.getName());
			toReturn = remotedTargetJavaInterface;
		}

		return toReturn;
	}

	/**
	 * if the interface of the component whose serviceBindings must be exposed
	 * as RMI Service, does not implement java.rmi.Remote, then generate such an
	 * interface. This method will stop with just generating the bytecode.
	 * Defining the class from the byte code must be the responsibility of the
	 * caller of this method, since it requires a ClassLoader to be created to
	 * define and load this interface.
	 */
	protected byte[] generateRemoteInterface(Class<?> serviceInterface) {
		String interfazeName = serviceInterface.getCanonicalName();
		ClassWriter cw = new ClassWriter(ClassWriter.COMPUTE_MAXS);

		final String internalNameGeneratedClass = interfazeName.replace('.',
				'/')
				+ GENERATED_CLASS_SUFFIX;
		log.finest("Internal name for generated class: "
				+ internalNameGeneratedClass);
		String[] implementedInterfaces = computeInterfacesfrom(serviceInterface);
		cw.visit(Opcodes.V1_5, Opcodes.ACC_PUBLIC + Opcodes.ACC_ABSTRACT
				+ Opcodes.ACC_INTERFACE, internalNameGeneratedClass, null,
				"java/lang/Object", implementedInterfaces);

		StringBuffer argsAndReturn = null;
		Method[] methods = serviceInterface.getMethods();
		for (Method method : methods) {
			argsAndReturn = new StringBuffer("(");
			Class<?>[] paramTypes = method.getParameterTypes();
			Class<?> returnType = method.getReturnType();

			for (Class<?> paramType : paramTypes) {
				argsAndReturn.append(Type.getType(paramType));
			}
			argsAndReturn.append(")");
			argsAndReturn.append(Type.getType(returnType));

			cw.visitMethod(Opcodes.ACC_PUBLIC + Opcodes.ACC_ABSTRACT, method
					.getName(), argsAndReturn.toString(), null,
					new String[] { "java/rmi/RemoteException" });
		}
		cw.visitEnd();
		return cw.toByteArray();
	}

	/**
	 * @param serviceInterface
	 * @return
	 */
	private String[] computeInterfacesfrom(Class<?> serviceInterface) {
		String[] itfs;
		final boolean isNotSerializable = !Serializable.class
				.isAssignableFrom(serviceInterface);
		if (isNotSerializable) {
			itfs = new String[3];
		} else {
			itfs = new String[2];
		}
		itfs[0] = "java/rmi/Remote";
		itfs[1] = Type.getInternalName(serviceInterface);
		if (isNotSerializable) {
			itfs[2] = "java/io/Serializable";
		}
		return itfs;
	}
}
