/**
 * Fractal Binding Factory.
 * Copyright (C) 2007-2008 INRIA, SARDES

 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contact: Valerio.Schiavoni@gmail.com
 * 
 * Author: Valerio Schiavoni
 * 
 */
package org.objectweb.fractal.bf.connectors.rmi.asm;

import java.lang.reflect.Method;
import java.util.logging.Logger;

import org.objectweb.asm.ClassWriter;
import org.objectweb.asm.FieldVisitor;
import org.objectweb.asm.Label;
import org.objectweb.asm.MethodVisitor;
import org.objectweb.asm.Opcodes;
import org.objectweb.asm.Type;

/**
 * Generate a class that implements the generated subclass of
 * {@link java.rmi.Remote}.
 */
public class RemoteSubclassGenerator implements Opcodes {

	/**
	 * The name of the field to generate referencing the object to which
	 * delegate calls.
	 */
	private static final String DELEGATED_FIELD_NAME = "s";
	Logger log = Logger.getLogger(RemoteSubclassGenerator.class
			.getCanonicalName());

	/**
	 * @param the
	 *            input class
	 */
	public Class<?> generate(Class<?> unremotableInterfaze,
			Class<?> remotableInterfaze) {
		final String unremotableInterfazeInternalName = Type
				.getInternalName(unremotableInterfaze);
		final String remotableInterfazeInternalName = Type
				.getInternalName(remotableInterfaze);

		log.finest("Internal type of delegate field: "
				+ unremotableInterfazeInternalName);

		ClassWriter cw = new ClassWriter(ClassWriter.COMPUTE_MAXS);
		FieldVisitor fv;
		MethodVisitor mv;

		final String generatedClassInternalName = remotableInterfazeInternalName
				+ "Impl";
		log.finest("Generated class internal name: "
				+ generatedClassInternalName);
		cw.visit(V1_5, ACC_PUBLIC + ACC_SUPER, generatedClassInternalName,
				null, "java/lang/Object",
				new String[] { remotableInterfazeInternalName,
						"java/io/Serializable" });

		{

			fv = cw.visitField(ACC_PUBLIC, DELEGATED_FIELD_NAME, "L"
					+ unremotableInterfazeInternalName + ";", null, null);
			fv.visitEnd();
			log.finest("Field name '" + DELEGATED_FIELD_NAME + "' of type "
					+ unremotableInterfazeInternalName + " generated");
		}

		{
			mv = cw.visitMethod(ACC_PUBLIC, "<init>", "(L"
					+ unremotableInterfazeInternalName + ";)V", null,
					new String[] { "java/rmi/RemoteException" });
			mv.visitCode();
			Label l0 = new Label();
			mv.visitLabel(l0);
			mv.visitVarInsn(ALOAD, 0);
			mv.visitMethodInsn(INVOKESPECIAL, "java/lang/Object", "<init>",
					"()V");
			Label l1 = new Label();
			mv.visitLabel(l1);
			mv.visitVarInsn(ALOAD, 0);
			mv.visitVarInsn(ALOAD, 1);
			mv.visitFieldInsn(PUTFIELD, generatedClassInternalName,
					DELEGATED_FIELD_NAME, "L"
							+ unremotableInterfazeInternalName + ";");
			Label l2 = new Label();
			mv.visitLabel(l2);
			mv.visitInsn(RETURN);
			Label l3 = new Label();
			mv.visitLabel(l3);
			mv.visitLocalVariable("this", "L" + generatedClassInternalName
					+ ";", null, l0, l3, 0);
			mv.visitLocalVariable(DELEGATED_FIELD_NAME, "L"
					+ unremotableInterfazeInternalName + ";", null, l0, l3, 1);
			mv.visitMaxs(2, 2);
			mv.visitEnd();
			log.fine("1-Args constructor generated.");

		}

		StringBuffer argsAndReturn = null;
		for (Method method : remotableInterfaze.getMethods()) {
			log.finest("Generating method body for " + method.getName());
			argsAndReturn = new StringBuffer("(");
			Class<?>[] paramTypes = method.getParameterTypes();
			Class<?> returnType = method.getReturnType();
			int args = 0;
			for (Class<?> paramType : paramTypes) {
				argsAndReturn.append(Type.getType(paramType));
				args++;
			}
			log.finest("Method has " + args + " arguments");
			argsAndReturn.append(")");
			argsAndReturn.append(Type.getType(returnType));

			mv = cw.visitMethod(Opcodes.ACC_PUBLIC, method.getName(),
					argsAndReturn.toString(), null,
					new String[] { "java/rmi/RemoteException" });

			mv.visitCode();

			Label l0 = new Label();
			mv.visitLabel(l0);
			mv.visitVarInsn(ALOAD, 0);

			mv.visitFieldInsn(GETFIELD, generatedClassInternalName,
					DELEGATED_FIELD_NAME, "L"
							+ unremotableInterfazeInternalName + ";");

			// for every parameter of this method, push the actual argument on
			// the stack
			int i = 1;
			while (i <= args) {
				mv.visitVarInsn(ALOAD, i);
				log.finest("ALOAD index: " + i);
				i = i + 1;
			}

			mv.visitMethodInsn(INVOKEINTERFACE,
					unremotableInterfazeInternalName, method.getName(),
					argsAndReturn.toString());

			final String returnTypeValue = returnType.toString();
			if (returnTypeValue.equalsIgnoreCase("void")) {
				log.finest("Method " + method
						+ " returns void, using RETURN opcode");
				mv.visitInsn(RETURN);
			} else {
				log.finest("Method " + method + " return " + returnTypeValue
						+ ", using ARETURN opcode");
				mv.visitInsn(ARETURN);
			}

			Label l1 = new Label();
			mv.visitLabel(l1);
			mv.visitLocalVariable("this", "L" + generatedClassInternalName
					+ ";", null, l0, l1, 0);

			mv.visitMaxs(1, 1);
			mv.visitEnd();

			log.finer("Method " + method.toGenericString() + " generated");
		}

		cw.visitEnd();
		byte[] bytecode = cw.toByteArray();

		RMIServiceClassLoader classloader = RMIServiceClassLoaderFactory
				.getRMIClassLoader();
		log.fine("RMIClassLoader retrieved, ref:" + classloader);
		log.finest("Classes defined by the RMIServiceClassLoader: "
				+ classloader.getDefinedClasses());
		final Class<?> definedClazz = classloader.defineClass(bytecode,
				generatedClassInternalName.replace("/", "."));
		log.finest("Class defined in RMIServiceClassLoader: " + definedClazz);
		log
				.finest("Classes defined by the RMIServiceClassLoader after the definition of the new unicast object: "
						+ classloader.getDefinedClasses());
		return definedClazz;

	}
}
